from ..const import PAYLOAD_DATA
from .transaction import Transaction
from .responses import PipelineOKResponse, PluginConfigOKResponse, PluginInstanceCommandOKResponse
from time import time, sleep


class Instance():
  """
  The Instance class is a wrapper around a plugin instance. It provides a simple API for sending commands to the instance and updating its configuration.
  """

  def __init__(
    self, 
    log, 
    pipeline, 
    instance_id, 
    signature, 
    on_data=None, 
    on_notification=None, 
    config={}, 
    is_attached=False, 
    debug=False,
    **kwargs
  ):
    """
    Create a new instance of the plugin.

    Parameters
    ----------
    log : Logger
        The logger object
    pipeline : Pipeline
        The pipeline that the instance is part of
    instance_id : str
        The unique identifier of the instance
    signature : str
        The name of the plugin signature
    on_data : Callable[[Pipeline, str, str, dict], None], optional
        Callback that handles messages received from this instance. As arguments, it has a reference to this Pipeline object, along with the payload itself.
        Defaults to None.
    on_notification : Callable[[Pipeline, dict], None], optional
        Callback that handles notifications received from this instance. As arguments, it has a reference to this Pipeline object, along with the payload itself.
        Defaults to None.
    config : dict, optional
        Parameters used to customize the functionality. One can change the AI engine used for object detection,
        or finetune alerter parameters to better fit a camera located in a low light environment.
        Defaults to {}
    """
    self.log = log
    self.pipeline = pipeline
    self.instance_id = instance_id
    self.last_known_status = None
    self.signature = signature.upper()
    self.config = {}
    self.__debug = debug

    if is_attached:
      assert len(kwargs) == 0, "When attaching an instance, no additional parameters are allowed"
      self.config = config
      self.proposed_config = None
    else:
      self.proposed_config = {**config, **kwargs}
      self.proposed_config = {k.upper(): v for k, v in self.proposed_config.items()}
    self.__staged_config = None
    self.__was_last_operation_successful = None

    self.on_data_callbacks = []
    self.temporary_on_data_callbacks = {}
    self.on_notification_callbacks = []
    self.temporary_on_notification_callbacks = {}

    if on_data:
      self.on_data_callbacks.append(on_data)
    if on_notification:
      self.on_notification_callbacks.append(on_notification)

    return
  
  def Pd(self, *args, **kwargs):
    if self.__debug:
      kwargs["color"] = 'd' if kwargs.get("color") != 'r' else 'r'
      self.log.P(*args, **kwargs)
    return

  # Message handling
  if True:
    def _on_data(self, pipeline, data):
      """
      Handle the data received from the instance.

      Parameters
      ----------
      pipeline : Pipeline
          The pipeline that the instance is part of
      data : dict | Payload
          The data received from the instance
      """
      self.Pd(f"_on_data: {pipeline.name}:{self.signature}:{self.instance_id}")
      for callback in self.on_data_callbacks:
        callback(pipeline, data)
      for callback in self.temporary_on_data_callbacks.values():
        callback(pipeline, data)
      return

    def _on_notification(self, pipeline, data):
      """
      Handle the notification received from the instance.

      Parameters
      ----------
      pipeline : Pipeline
          The pipeline that the instance is part of
      data : dict | Payload
          The notification received from the instance
      """
      for callback in self.on_notification_callbacks:
        callback(pipeline, data)
      for callback in self.temporary_on_notification_callbacks.values():
        callback(pipeline, data)
      return

    def _add_on_data_callback(self, callback):
      """
      Add a new callback to the list of callbacks that handle the data received from the instance.

      Parameters
      ----------
      callback : Callable[[Pipeline, dict], None]
          The callback to add
      """
      self.on_data_callbacks.append(callback)
      return

    def _add_temporary_on_data_callback(self, attachment, callback):
      """
      Add a new temporary callback to the list of callbacks that handle the data received from the instance.

      Parameters
      ----------
      attachment : object
          The attachment id of the callback
      callback : Callable[[Pipeline, dict], None]
          The callback to add
      """
      # TODO: this can fail (very small chance, but still)
      # FIXME: make add / delete happen after callbacks
      self.temporary_on_data_callbacks[attachment] = callback
      return

    def _remove_temporary_on_data_callback(self, attachment):
      """
      Remove a temporary callback from the list of callbacks that handle the data received from the instance.

      Parameters
      ----------
      attachment : object
          The attachment id of the callback
      """
      if attachment in self.temporary_on_data_callbacks:
        del self.temporary_on_data_callbacks[attachment]
      return

    def _reset_on_data_callback(self):
      """
      Reset the list of callbacks that handle the data received from the instance.
      """
      self.on_data_callbacks = []
      return

    def _add_on_notification_callback(self, callback):
      """
      Add a new callback to the list of callbacks that handle the notifications received from the instance.

      Parameters
      ----------
      callback : Callable[[Pipeline, dict], None]
          The callback to add
      """
      self.on_notification_callbacks.append(callback)
      return

    def _add_temporary_on_notification_callback(self, attachment, callback):
      """
      Add a new temporary callback to the list of callbacks that handle the notifications received from the instance.

      Parameters
      ----------
      attachment : object
          The attachment id of the callback
      callback : Callable[[Pipeline, dict], None]
          The callback to add
      """
      self.temporary_on_notification_callbacks[attachment] = callback
      return

    def _remove_temporary_on_notification_callback(self, attachment):
      """
      Remove a temporary callback from the list of callbacks that handle the notifications received from the instance.

      Parameters
      ----------
      attachment : object
          The attachment id of the callback
      """
      if attachment in self.temporary_on_notification_callbacks:
        del self.temporary_on_notification_callbacks[attachment]
      return

    def _reset_on_notification_callback(self):
      """
      Reset the list of callbacks that handle the notifications received from the instance.
      """
      self.on_notification_callbacks = []
      return

  # Utils
  if True:
    def __repr__(self) -> str:
      node_addr = self.pipeline.node_addr
      pipeline_name = self.pipeline.name
      signature = self.signature
      instance_id = self.instance_id

      return f"<Instance: {node_addr}/{pipeline_name}/{signature}/{instance_id}>"


    def _is_tainted(self):
      """
      Check if the instance has a proposed configuration.

      Returns
      -------
      bool
          True if the instance has a proposed configuration, False otherwise
      """
      return self.proposed_config is not None


    def _get_config_dictionary(self):
      """
      Get the configuration of the instance as a dictionary.

      Returns
      -------
      dict
          The configuration of the instance as a dictionary
      """
      config_dict = {
        'INSTANCE_ID': self.instance_id,
        **self.config
      }

      return config_dict


    def _get_proposed_config_dictionary(self, full=False):
      """
      Get the proposed configuration of the instance as a dictionary.

      Returns
      -------
      dict
          The proposed configuration of the instance as a dictionary
      """
      if self.proposed_config is None:
        return self._get_config_dictionary()

      proposed_config_dict = {
        'INSTANCE_ID': self.instance_id,
        **({} if not full else self.config),
        **self.proposed_config
      }

      return proposed_config_dict


    def _apply_staged_config(self, verbose=False):
      """
      Apply the staged configuration to the instance.
      """
      if self.__staged_config is None:
        return

      if verbose:
        self.P(f'Applying staged configuration to instance <{self.instance_id}>', color="g")
      self.__was_last_operation_successful = True

      self.config = {**self.config, **self.__staged_config}
      self.__staged_config = None
      return


    def _discard_staged_config(self, fail_reason: str):
      """
      Discard the staged configuration for the instance.
      """
      self.P(f'Discarding staged configuration for instance <{self.instance_id}>. Reason: {fail_reason}', color="r")
      self.__was_last_operation_successful = False

      self.__staged_config = None
      return


    def _stage_proposed_config(self):
      """
      Stage the proposed configuration for the instance.
      """
      if self.proposed_config is None:
        return

      if self.__staged_config is not None:
        raise ValueError("Instance configuration has already been staged, waiting for confirmation from Execution Engine")

      # self.__staged_config is None
      self.__staged_config = self.proposed_config
      self.proposed_config = None

      self.__was_last_operation_successful = None
      return


    def _handle_instance_command_success(self):
      """
      Handle the success of the instance command.
      """
      self.P(f'Instance command successful for instance <{self.instance_id}>', color="g")
      self.__was_last_operation_successful = True
      return


    def _handle_instance_command_failure(self, fail_reason: str):
      """
      Handle the failure of the instance command.
      """
      self.P(f'Instance command failed for instance <{self.instance_id}>. Reason: {fail_reason}', color="r")
      self.__was_last_operation_successful = False
      return


    def __register_transaction_for_instance_command(self, session_id: str = None, timeout: float = 0) -> list[Transaction]:
      """
      Register a new transaction for the instance command.
      This method is called before sending an instance command to the ratio1 Edge Protocol edge node.

      Parameters
      ----------
      session_id : str, optional
          The session ID of the transaction, by default None
      timeout : float, optional
          The timeout for the transaction, by default 0

      Returns
      -------
      list[Transaction]
          The list of transactions generated
      """
      required_responses = [
        PipelineOKResponse(self.pipeline.node_id, self.pipeline.name),
        PluginInstanceCommandOKResponse(self.pipeline.node_id, self.pipeline.name, self.signature, self.instance_id),
        # PluginConfigOKResponse(self.pipeline.node_id, self.pipeline.name, self.signature, self.instance_id),
      ]

      transactions = [self.pipeline.session._register_transaction(
          session_id=session_id,
          lst_required_responses=required_responses,
          timeout=timeout,
          on_success_callback=self._handle_instance_command_success,
          on_failure_callback=self._handle_instance_command_failure,
      )]

      return transactions


    def _get_instance_update_required_responses(self):
      """
      Get the responses required to update the instance.

      Returns
      -------
      list[str]
          The list of responses required to update the instance
      """
      responses = [
        PipelineOKResponse(self.pipeline.node_id, self.pipeline.name),
        PluginConfigOKResponse(self.pipeline.node_id, self.pipeline.name, self.signature, self.instance_id),
      ]

      return responses


    def _get_instance_remove_required_responses(self):
      """
      Get the responses required to delete the instance.

      Returns
      -------
      list[str]
          The list of responses required to delete the instance
      """
      responses = [
        PipelineOKResponse(self.pipeline.node_id, self.pipeline.name),
      ]

      return responses

    def _get_method_data(self, method: callable):
      """
      Get the full signature and definition of a method.
      
      Parameters
      ----------
      
      method : callable
          The method to get the signature and definition for.
          
      Returns
      -------
      tuple
          A tuple containing the name, arguments and base64 code of the method.
      """
      return self.pipeline._get_method_data(method)
    
  # API
  if True:
    @property
    def was_last_operation_successful(self) -> bool:
      """
      Get the status of the last operation.

      Example:
      --------
      ```python
      instance.send_instance_command("RESTART", wait_confirmation=False)

      if instance.was_last_operation_successful is not None:
        if instance.was_last_operation_successful:
          print("Last operation was successful")
        else:
          print("Last operation failed")
      else:
        print("No ACK received yet for this operation")
      ```

      Returns
      -------
      bool, None
          True if the last operation was successful, False if it failed, None if the ACK has not been received yet
      """
      return self.__was_last_operation_successful


    def _sync_configuration_with_remote(self, config):
      self.config = {**self.config, **config}
      return


    def update_instance_config(self, config={}, **kwargs):
      """
      Update the configuration of the instance.
      The new configuration is merged with the existing one.
      Parameters can be passed as a dictionary in `config` or as `kwargs`.

      Parameters
      ----------
      config : dict, optional
          The new configuration of the instance, by default {}

      Returns
      -------
      dict | None
          The new configuration of the instance as a dictionary if `send_command` is False, otherwise None
      """

      if self.__staged_config is not None:
        raise ValueError("Instance configuration has already been staged, waiting for confirmation from Execution Engine")

      if self.proposed_config is None:
        self.proposed_config = {}

      self.proposed_config = {**self.proposed_config, **config, **{k.upper(): v for k, v in kwargs.items()}}

      for k, v in self.config.items():
        if k in self.proposed_config:
          if self.proposed_config[k] == v:
            del self.proposed_config[k]

      if len(self.proposed_config) == 0:
        self.proposed_config = None

      return


    def send_instance_command(self, command, payload=None, command_params=None, wait_confirmation=True, session_id=None, timeout=10):
      """
      Send a command to the instance.
      This command can block until the command is confirmed by the ratio1 Edge Protocol edge node.

      Example:
      --------
      ```python
      instance.send_instance_command('START', wait_confirmation=True)

      transactions_p1 = instance1.send_instance_command('START', wait_confirmation=False)
      transactions_p2 = instance2.send_instance_command('START', wait_confirmation=False)
      # wait for both commands to be confirmed, but after both commands are sent
      session.wait_for_transactions(transactions_p1 + transactions_p2)
      ```
      Parameters
      ----------
      command : str
          The command to send
      payload : dict, optional
          The payload of the command, by default {}
      command_params : dict, optional
          The parameters of the command, by default {}
      wait_confirmation : bool, optional
          Whether to wait for the confirmation of the command, by default False
      timeout : int, optional
          The timeout for the transaction, by default 10    

      Returns
      -------
      list[Transaction] | None
          The list of transactions generated, or None if `wait_confirmation` is False.

      """
      self.Pd(f'Sending command <{command}> to instance <{self.__repr__()}>', color="b")

      self.__was_last_operation_successful = None

      transactions = self.__register_transaction_for_instance_command(timeout=timeout)

      self.pipeline.session._send_command_instance_command(
        worker=self.pipeline.node_addr,
        pipeline_name=self.pipeline.name,
        signature=self.signature,
        instance_id=self.instance_id,
        command=command,
        payload=payload,
        command_params=command_params,
        session_id=session_id,
      )

      if wait_confirmation:
        self.pipeline.session.wait_for_transactions(transactions)
      else:
        return transactions
      return


    def close(self):
      """
      Close the instance.
      """
      self.pipeline.remove_plugin_instance(self)
      return


    def stop(self):
      """
      Close the instance. Alias for `close`.
      """
      self.close()


    def P(self, *args, **kwargs):
      self.log.P(*args, **kwargs)
      return


    def D(self, *args, **kwargs):
      self.log.D(*args, **kwargs)
      return


    def temporary_attach(self, on_data=None, on_notification=None):
      """
      Attach a temporary callback to the instance.

      Parameters
      ----------
      on_data : Callable[[Pipeline, str, str, dict], None], optional
          Callback that handles messages received from this instance. As arguments, it has a reference to this Pipeline object, along with the payload itself.
          Defaults to None.
      on_notification : Callable[[Pipeline, dict], None], optional
          Callback that handles notifications received from this instance. As arguments, it has a reference to this Pipeline object, along with the payload itself.
          Defaults to None.

      Returns
      -------
      object
          The attachment id of the callback
      """
      attachment = object()
      if on_data:
        self._add_temporary_on_data_callback(attachment, on_data)
      if on_notification:
        self._add_temporary_on_notification_callback(attachment, on_notification)

      return attachment


    def temporary_detach(self, attachment):
      """
      Detach a temporary callback from the instance.

      Parameters
      ----------
      attachment : object
          The attachment id of the callback
      """
      self._remove_temporary_on_data_callback(attachment)
      self._remove_temporary_on_notification_callback(attachment)
      return

   
    def convert_to_specialized_class(self, specialized_class):
      """
      Convert the object to a specialized class.
      A specialized class is a class that inherits from the Instance class and 
      provides additional methods for ease of use.
      """
      self.__class__ = specialized_class
      return self


    def send_instance_command_and_wait_for_response_payload(self, command, payload=None, command_params=None, timeout_command=10, timeout_response_payload=3, response_params_key="COMMAND_PARAMS"):
      """
      Send a command to the instance and wait for the response payload.

      Parameters
      ----------
      command : str | dict
          The command to send
      payload : dict, optional
          The payload of the command, by default {}
      command_params : dict, optional
          The parameters of the command, by default {}
      timeout : int, optional
          The timeout for the transaction, by default 10

      Returns
      -------
      dict: dict | None
          The payload received from the instance, or None if the command failed or if the payload was not received
      """
      result_payload = None
      uid = self.log.get_uid()

      def wait_payload_on_data(pipeline, data):
        nonlocal result_payload
        if response_params_key in data and data[response_params_key].get("SDK_REQUEST") == uid:
          result_payload = data
        return

      attachment = self.temporary_attach(on_data=wait_payload_on_data)

      if payload is None:
        payload = {}
      payload["SDK_REQUEST"] = uid

      self.send_instance_command(
        command=command,
        payload=payload,
        command_params=command_params,
        wait_confirmation=True,
        timeout=timeout_command,
      )

      start_time = time()
      while time() - start_time < timeout_response_payload and result_payload is None:
        sleep(0.1)

      self.temporary_detach(attachment)

      return result_payload

  def get_status(self):
    """
    Get the status of the instance.
    
    Returns
    -------
    dict
        The status of the instance with keys from `PLUGIN_INFO` class constants
                
    Example:
    --------
    ```json
    {
      "CURRENT_EXEC_ITERATION": 2147088,
      "CURRENT_PROCESS_ITERATION": 0,
      "EXEC_TIMESTAMP": "2025-02-28 07:06:40.612999",
      "FIRST_ERROR_TIME": null,
      "FREQUENCY": null,
      "INFO": null,
      "INIT_TIMESTAMP": "2025-02-27 10:24:56.622480",
      "INSTANCE_ID": "RECE01_default",
      "LAST_CONFIG_TIMESTAMP": "2025-02-27 10:24:56.622340",
      "LAST_ERROR_TIME": null,
      "LAST_PAYLOAD_TIME": "2025-02-27 10:24:56",
      "OUTSIDE_WORKING_HOURS": false,
      "PROCESS_DELAY": 1,
      "SIGNATURE": "REST_CUSTOM_EXEC_01",
      "STREAM_ID": "admin_pipeline",
      "TOTAL_PAYLOAD_COUNT": 1
    }
    ```    
    """
    result = {} if self.last_known_status is None else self.last_known_status
    if self.last_known_status is None:
      self.Pd(f'Instance <{self.instance_id}> has no status yet')
    return result


import os
import base64

class _ChainMixin:
  def create_r1_token(self) -> str:
    """
    Create an online-usable token with the format "r1_<encoded>"
    that encapsulates both the SDK address (omitting '0x') and a nonce.
    """
    # Encode the SDK address (sans '0x') as UTF-8
    addr_bytes = self.address_no_prefix.encode('utf-8')

    # Generate an 8-byte random nonce
    nonce_bytes = os.urandom(8)

    # Add a delimiter (0x00) between address and nonce
    raw_data = addr_bytes + b'\x00' + nonce_bytes

    # Base64-encode the combined data
    encoded_part = base64.urlsafe_b64encode(raw_data).decode('utf-8')
    return f"r1_{encoded_part}"

  def check_r1_token(self, token: str, return_nonce: bool = False):
    """
    Verify the token’s format ("r1_<encoded>"), extract address and nonce.

    Parameters
    ----------
    token : str
      The token string (must start with "r1_").

    return_nonce : bool, optional
      If True, return (address, nonce). If False, return only address.

    Returns
    -------
    str or tuple
      - If return_nonce is False, returns the address (with "0x" prefix if needed).
      - If return_nonce is True, returns (address_str, nonce_hex).

    Raises
    ------
    ValueError
      If the token is malformed or doesn't start with "r1_".
    """
    if not token.startswith("r1_"):
      raise ValueError("Invalid token format. Missing 'r1_' prefix.")

    encoded_part = token[len("r1_"):]
    try:
      raw_data = base64.urlsafe_b64decode(encoded_part.encode('utf-8'))
    except Exception:
      raise ValueError("Invalid token: base64 decoding failed.")

    # Find our delimiter (0x00) separating address bytes from nonce
    delimiter_index = raw_data.find(b'\x00')
    if delimiter_index == -1 or delimiter_index == len(raw_data) - 1:
      raise ValueError("Invalid token structure: missing or misplaced delimiter.")

    addr_bytes = raw_data[:delimiter_index]
    nonce_bytes = raw_data[delimiter_index + 1:]

    # Convert the address bytes back to string. Optionally re-add "0x" if your logic requires.
    address_str = addr_bytes.decode('utf-8')
    address_str = self.maybe_add_prefix(address_str)

    # Convert the nonce to a hex string
    nonce_hex = nonce_bytes.hex()

    if return_nonce:
      return address_str, nonce_hex
    return address_str



from ...base import Instance
from ...const import PLUGIN_SIGNATURES

class ViewScene01(Instance):
  signature = PLUGIN_SIGNATURES.VIEW_SCENE_01

  def get_last_witness(self, response_params_key="COMMAND_PARAMS"):
    """
    Get the performance history of a node.

    Parameters
    ----------
    response_params_key : str
      The key in the response payload that contains the command parameters
    Returns
    -------
    """
    images = []

    result_payload = self.send_instance_command_and_wait_for_response_payload(
      command="GET_LAST_WITNESS",
      response_params_key=response_params_key,
    )

    if result_payload is not None:
      images = result_payload.get_images_as_PIL()

    return images



import os
import time

from ratio1 import Session, CustomPluginTemplate

def reply(plugin, message: str, user: str, **kwargs):
  # Define the deck of cards and their values
  cards = ['2', '3', '4', '5', '6', '7', '8', '9', '10',
           'Jack', 'Queen', 'King', 'Ace']
  card_values = {
    '2': 2, '3': 3, '4': 4, '5': 5, '6': 6, '7': 7,
    '8': 8, '9': 9, '10': 10,
    'Jack': 10, 'Queen': 10, 'King': 10,
    'Ace': 11  # Ace is treated as 11 initially; we'll adjust as needed
  }

  # Initialize user state if not already present
  if user not in plugin.obj_cache or plugin.obj_cache[user] is None:
    plugin.obj_cache[user] = {
      'state': 'NOT_STARTED',
      'wins': 0,
      'losses': 0
    }

  # Retrieve the user's game state
  user_cache = plugin.obj_cache[user]
  message = message.strip().lower()  # Normalize the input message

  if user_cache.get('state') == 'NOT_STARTED':
    if message == 'card' or message == 'start':
      # Start a new game
      user_cache['hand'] = []  # Initialize player's hand
      user_cache['state'] = 'PLAYING'

      # Deal the first card
      card = plugin.np.random.choice(cards)
      user_cache['hand'].append(card)

      # Calculate hand value
      total = 0
      aces = 0  # Count of aces in hand
      for c in user_cache['hand']:
        if c == 'Ace':
          aces += 1
        else:
          total += card_values[c]
      # Adjust for aces
      for _ in range(aces):
        if total + 11 <= 21:
          total += 11  # Count ace as 11 if it doesn't cause bust
        else:
          total += 1   # Otherwise, count ace as 1

      # Return the initial game message
      return (f"Welcome to a new game of Black Jack, you have a {card}. "
              f"Your total is {total}. Type 'card' to get another card, or 'stop' to hold.")
    else:
      # Prompt user to start a new game
      return "Welcome to Black Jack. Type 'card' to start a new game."

  elif user_cache.get('state') == 'PLAYING':
    if message == 'card':
      # Player chooses to draw another card
      card = plugin.np.random.choice(cards)
      user_cache['hand'].append(card)

      # Recalculate hand value
      total = 0
      aces = 0
      for c in user_cache['hand']:
        if c == 'Ace':
          aces += 1
        else:
          total += card_values[c]
      for _ in range(aces):
        if total + 11 <= 21:
          total += 11
        else:
          total += 1

      hand_str = ', '.join(user_cache['hand'])  # Format hand as a string

      if total > 21:
        # Player busts
        user_cache['state'] = 'BUST'
        user_cache['losses'] += 1  # Increment losses
        return (f"You are given a {card}. Your hand: {hand_str}. "
                f"Your total is {total}. Bust! You lose.\n"
                f"Total Wins: {user_cache['wins']}, Total Losses: {user_cache['losses']}")
      elif total == 21:
        # Player hits Black Jack
        user_cache['state'] = 'BLACKJACK'
        user_cache['wins'] += 1  # Increment wins
        return (f"You are given a {card}. Black Jack! Congratulations, you win! "
                f"Your hand: {hand_str}.\n"
                f"Total Wins: {user_cache['wins']}, Total Losses: {user_cache['losses']}")
      else:
        # Continue playing
        return (f"You are given a {card}. Your hand: {hand_str}. "
                f"Your total is {total}. Type 'card' to get another card, or 'stop' to hold.")
    elif message == 'stop':
      # Player chooses to hold; dealer's turn
      dealer_hand = []
      dealer_total = 0
      dealer_aces = 0

      # Dealer draws cards until total >= 17
      while True:
        dealer_card = plugin.np.random.choice(cards)
        dealer_hand.append(dealer_card)

        # Update dealer's total
        if dealer_card == 'Ace':
          dealer_aces += 1
        else:
          dealer_total += card_values[dealer_card]

        # Calculate dealer's total considering aces
        temp_total = dealer_total
        aces = dealer_aces
        for _ in range(aces):
          if temp_total + 11 <= 21:
            temp_total += 11
          else:
            temp_total += 1

        dealer_total_with_aces = temp_total

        # Dealer stops hitting when total >= 17
        if dealer_total_with_aces >= 17:
          dealer_total = dealer_total_with_aces
          break
        else:
          dealer_total = dealer_total_with_aces

      # Calculate player's total
      player_total = 0
      player_aces = 0
      for c in user_cache['hand']:
        if c == 'Ace':
          player_aces += 1
        else:
          player_total += card_values[c]
      for _ in range(player_aces):
        if player_total + 11 <= 21:
          player_total += 11
        else:
          player_total += 1

      user_cache['state'] = 'GAME_OVER'  # Update game state

      # Determine the outcome
      if dealer_total > 21:
        result = "Dealer busts! You win!"
        user_cache['wins'] += 1
      elif player_total > dealer_total:
        result = "You win!"
        user_cache['wins'] += 1
      elif player_total == dealer_total:
        result = "It's a tie!"
        # Ties are not counted as wins or losses
      else:
        result = "You lose!"
        user_cache['losses'] += 1

      player_hand_str = ', '.join(user_cache['hand'])
      dealer_hand_str = ', '.join(dealer_hand)

      # Return the game result with total wins and losses
      return (f"You stopped with {player_total}. Your hand: {player_hand_str}.\n"
              f"Dealer's hand: {dealer_hand_str} (Total: {dealer_total}).\n"
              f"{result}\nTotal Wins: {user_cache['wins']}, Total Losses: {user_cache['losses']}")
    else:
      # Invalid input during playing
      return "Invalid input. Please type 'card' to get another card, or 'stop' to hold."

  elif user_cache.get('state') in ['BUST', 'BLACKJACK', 'GAME_OVER']:
    # Game is over; reset state for a new game
    user_cache['state'] = 'NOT_STARTED'
    return "Game over. Type 'card' to start a new game."

  else:
    # Catch-all for unexpected state
    return "An error occurred. Please start a new game."



if __name__ == "__main__":
    
  session = Session() 
  
  # assume .env is available and will be used for the connection and tokens
  # NOTE: When working with SDK please use the nodes internal addresses. While the EVM address of the node
  #       is basically based on the same sk/pk it is in a different format and not directly usable with the SDK
  #       the internal node address is easily spoted as starting with 0xai_ and can be found 
  #       via `docker exec r1node get_node_info` or via the launcher UI
  my_node = os.getenv("EE_TARGET_NODE", "0xai_your_node_address") # we can specify a node here, if we want to connect to a specific

  session.wait_for_node(my_node) # wait for the node to be active
    
  # unlike the previous example, we are going to use the token from the environment
  # and deploy the app on the target node and leave it there
  pipeline, _ = session.create_telegram_simple_bot(
    node=my_node,
    name="telegram_bot_blackjack",
    message_handler=reply,
  )
  
  pipeline.deploy() # we deploy the pipeline


  # Observation:
  #   next code is not mandatory - it is used to keep the session open and cleanup the resources
  #   due to the fact that this is a example/tutorial and maybe we dont want to keep the pipeline
  #   active after the session is closed we use close_pipelines=True
  #   in production, you would not need this code as the script can close 
  #   after the pipeline will be sent 
  session.wait(
    seconds=600,            # we wait the session for 10 minutes
    close_pipelines=True,   # we close the pipelines after the session
    close_session=True,     # we close the session after the session
  )
 



import time
from .pipeline import Pipeline



class WebappPipeline(Pipeline):
  
  def __init__(
    self, 
    session, 
    log, 
    *, 
    node_addr, 
    name, 
    config={}, 
    plugins=[], 
    is_attached=False, 
    on_data=None,
    on_notification=None,
    existing_config=None, 
    extra_debug=False,
    **kwargs
  ) -> None:
    """
    This is a special type of pipeline that is used to deploy webapps.
    It will override the deploy method to return the app URL as well as the on_data method to extract the app URL.
    """
    self.app_url = None
    self.__extra_debug = extra_debug
    _on_data = [self.__check_payloads, on_data]    
    super().__init__(
      session, 
      log, 
      node_addr=node_addr, 
      name=name, 
      config=config, 
      plugins=plugins, 
      is_attached=is_attached, 
      on_data=_on_data,
      on_notification=on_notification,
      existing_config=existing_config, 
      
      **kwargs
    )
    return
  
  
  def __check_payloads(self, session, plugin_signature, plugin_instance, data):
    """
    Check if the payload is from webapp and extract the app URL.
    
    Parameters
    ----------
    
    
    session : Session
        The session object that received the payload.
            
    plugin_signature : str
        The signature of the plugin that sent the payload.
        
    plugin_instance : str
        The instance of the plugin that sent the payload.
        
    data : Payload
        The payload received from the edge node.     
        
         
    """
    if self.__extra_debug:
      self.P(f"Received payload from {plugin_signature} ({plugin_instance})")
    # TODO: this will be removed
    if "NGROK_URL" in data:
      self.app_url = data["NGROK_URL"]
    if "APP_URL" in data:
      self.app_url = data["APP_URL"]
    return
    
  def create_plugin_instance(
    self, 
    *, 
    signature, 
    instance_id, 
    config={},
    tunnel_engine='ngrok',
    tunnel_engine_enabled=True,
    ngrok_edge_label=None,
    cloudflare_token=None,
    **kwargs
  ):
    tunnel_kwargs = {
      'tunnel_engine': tunnel_engine,
      'tunnel_engine_enabled': tunnel_engine_enabled,
    }
    if not tunnel_engine_enabled:
      self.app_url = "TUNNEL_ENGINE_DISABLED"
    else:
      if tunnel_engine.lower() == 'cloudflare':
        self.P("Using Cloudflare as tunnel engine", color="green")
        if cloudflare_token is not None:
          self.app_url = "URL_DEFINED_IN_CLOUDFLARE_TOKEN"
          tunnel_kwargs['cloudflare_token'] = cloudflare_token
      else:
        self.P("Using ngrok as tunnel engine", color="green")
        if ngrok_edge_label is not None:
          self.app_url = "URL_DEFINED_IN_NGROK_EDGE_LABEL"
          tunnel_kwargs['ngrok_edge_label'] = ngrok_edge_label
        # endif ngrok_edge_label is not None
      # endif tunnel_engine ngrok or cloudflare
    # endif tunnel_engine_enabled
      
    return super().create_plugin_instance(
      signature=signature, 
      instance_id=instance_id, 
      config=config,
      **tunnel_kwargs,
      **kwargs
    )    
  
  def deploy(self, verbose=False, timeout=15, **kwargs):
    """
    Deploy the pipeline and return the URL of the webapp.
    
    Returns
    -------
    str
        The URL of the webapp.
    """
    res = super().deploy(verbose=verbose, timeout=timeout, **kwargs)
    # now we wait for the app url to be available
    start = time.time()
    while self.app_url is None:
      elapsed = time.time() - start
      if elapsed > timeout:
        msg = "Timeout waiting for app url"
        self.P(msg, color="red")
        raise Exception(msg)
    # return the app url
    return self.app_url


import json
import os
import queue
import threading
from datetime import time, datetime

from ratio1 import Session, Payload, PAYLOAD_DATA, HEARTBEAT_DATA


class MS_TYPE:
  HB = 'hb'
  PL = 'pl'
  NT = 'nt'


class NetworkSniffer:
  """
  A multithreaded network sniffer that collects and writes heartbeat, payload,
  and notification data from a `Session` to separate files in NDJSON format.
  It uses queues to manage incoming data and threads to write data to files
  asynchronously. The sniffer also logs its start and stop times and provides
  a summary of the collected data.
  Note: NDJSON (Newline Delimited JSON) is a format where each line
        in a file is a valid JSON object. It is commonly used for streaming data or
        storing large datasets because it allows processing one JSON object at a time
        without loading the entire file into memory.
  """
  def __init__(self, output_path: str = None, collection_period_seconds: int = 60):

    if output_path is None:
      date_started = datetime.now().isoformat()
      output_path = os.path.join('../network_sniffer/output', 'network_sniffer',
                                 date_started.replace(':', '-').replace('.', '-') + '_' + str(
                                   collection_period_seconds))
      os.makedirs(output_path, exist_ok=True)

    self.flush_every_n = 1

    self.notifications_queue = queue.Queue()
    self.heartbeats_queue = queue.Queue()
    self.payloads_queue = queue.Queue()

    self.output_path = output_path
    self.files = {
      MS_TYPE.HB: os.path.join(output_path, 'heartbeats.ndjson'),
      MS_TYPE.PL: os.path.join(output_path, 'payloads.ndjson'),
      MS_TYPE.NT: os.path.join(output_path, 'notifications.ndjson')
    }
    self._threads = []
    self.start()

  def on_heartbeat(self, session: Session, node_addr: str, heartbeat: dict):
    self.heartbeats_queue.put_nowait(heartbeat)
    return

  def on_payload(
      self,
      session: Session,
      node_addr: str,
      pipeline_name: str,
      plugin_signature: str,
      plugin_instance: str,
      payload: Payload
  ):
    self.payloads_queue.put_nowait(payload.data)
    return

  def on_notification(self, session, address, notification):
    self.notifications_queue.put_nowait(notification.data)
    return

  def start(self):
    self.date_started = datetime.now().isoformat()
    self.is_running = True

    self._threads.append(
      threading.Thread(target=self.writer_loop, args=(self.heartbeats_queue, MS_TYPE.HB), daemon=False))
    self._threads.append(
      threading.Thread(target=self.writer_loop, args=(self.payloads_queue, MS_TYPE.PL), daemon=False))
    self._threads.append(
      threading.Thread(target=self.writer_loop, args=(self.notifications_queue, MS_TYPE.NT), daemon=False))

    for thread in self._threads:
      thread.start()

  def stop(self):
    self.date_stopped = datetime.now().isoformat()
    self.is_running = False
    for thread in self._threads:
      thread.join()

  def writer_loop(self, q, ms_type):
    buffer = []
    filename = self.files[ms_type]
    first_item_written = False

    with open(filename, 'a') as f:
      f.seek(0, os.SEEK_END)

      try:
        while self.is_running or not q.empty():
          try:
            item = q.get(timeout=1)
            buffer.append(item)
            if len(buffer) >= self.flush_every_n:
              for item in buffer:
                try:
                  f.write(json.dumps(item) + '\n')
                except Exception as e:
                  print(f"Error writing {ms_type} data to file {filename}: {e}")
              buffer.clear()
              f.flush()
          except Exception as e:
            continue
          # end try get from queue
        # end while loop
      except Exception as e:
        print(f"Error in writer loop for {ms_type}: {e}")
      finally:
        if len(buffer):
          for item in buffer:
            try:
              f.write(json.dumps(item) + '\n')
            except Exception as e:
              print(f"Error writing {ms_type} data to file {filename}: {e}")
        f.flush()
    return

  def write_log_file(self):
    filename = os.path.join(self.output_path, f"network_sniffer_{self.date_started}.log")
    with open(filename, 'w') as f:
      f.write(f"Network Sniffer started at {self.date_started}\n")
      f.write(f"Network Sniffer stopped at {self.date_stopped}\n")
      f.write(f"Heartbeats written to: {self.files[MS_TYPE.HB]}\n")
      f.write(f"Payloads written to: {self.files[MS_TYPE.PL]}\n")
      f.write(f"Notifications written to: {self.files[MS_TYPE.NT]}\n")
      f.flush()
    return

def run_network_sniffer(collection_period_seconds: int = 60):
  date_started = datetime.now().isoformat()

  output_path = os.path.join('output', 'network_sniffer', date_started.replace(':', '-').replace('.', '-') + '_' + str(
    collection_period_seconds))
  os.makedirs(output_path, exist_ok=True)

  sniffer = NetworkSniffer(output_path, collection_period_seconds=collection_period_seconds)

  session = Session(
    on_heartbeat=sniffer.on_heartbeat,
    on_payload=sniffer.on_payload,
    on_notification=sniffer.on_notification,
    verbosity=3,
  )

  session.log.log_file = os.path.join(output_path, f"network_sniffer_{date_started}.log")

  session.P(
    f"Session started at {date_started}, collecting messages for {collection_period_seconds} seconds...")

  session.wait(seconds=collection_period_seconds, close_session=True)
  session.close()
  session.P("Session finished successfully!", color='g')
  sniffer.stop()
  sniffer.write_log_file()
  session.P("Script run successfully!")


if __name__ == '__main__':
  SNIFFER_COLLECTION_PERIOD_SECONDS = 1800
  run_network_sniffer(collection_period_seconds=SNIFFER_COLLECTION_PERIOD_SECONDS)



"""
This is a simple example of how to use the ratio1 SDK.

In this example, we connect to the network, choose a node and
    deploy a plugin with custom code that will run in real time.
    
For this example, we will count the number of persons that appear in the frames
    of a movie, so we will use a general object detection model.
"""

from time import sleep

from ratio1 import CustomPluginTemplate, Instance, Payload, Pipeline, Session


def real_time_code(plugin: CustomPluginTemplate):
  frame = plugin.dataapi_image()
  plugin.set_image(frame)

  persons = plugin.dataapi_image_instance_inferences()
  persons_count = len(persons)

  if persons_count > 0:
    return {
      'PERSONS_COUNT': persons_count,
    }

  return


def custom_instance_on_data(pipeline: Pipeline, custom_code_data: dict, data: Payload):
  # the images can be extracted from the Payload object
  # PIL needs to be installed for this to work
  data.get_images_as_PIL()[0].save('frame.jpg')
  pipeline.P("Persons in frame: {}".format(custom_code_data['PERSONS_COUNT']))


if __name__ == '__main__':
  # create a session
  # the network credentials are read from the .env file automatically
  session: Session = Session()

  # wait for any node to be available
  session.wait_for_any_node(timeout=10)

  # get the first available node
  chosen_node = session.get_active_nodes()[0]

  # we have our node, let's deploy a plugin

  # first, we create a pipeline
  # we will use the video file data source, since we want to extract frames from a video
  pipeline: Pipeline = session.create_pipeline(
    node=chosen_node,
    name='real_time_custom_code_deploy',
    data_source='VideoFile',
    config={
      'URL': "https://www.dropbox.com/scl/fi/8z2wpeelhav3k2dv8bb5p/Cars_3.mp4?rlkey=imv415rr3j1tx3zstpurlxkqb&dl=1"
    }
  )

  # next, we deploy a custom code plugin instance
  instance: Instance = pipeline.create_custom_plugin_instance(
    instance_id='inst01',
    custom_code=real_time_code,
    on_data=custom_instance_on_data,
    # we can specify the configuration for the plugin instance as kwargs
    process_delay=3,
    allow_empty_inputs=False,
    ai_engine="lowres_general_detector",
    object_type=["person"],
    # we can also specify if the payloads should be encrypted
    # if so, only the creator of this pipeline, in our case us, will be able to decrypt the payloads
    encrypt_payload=True,
  )

  # we increase the timeout because the AI engine has to start and warm up, which can take a while
  pipeline.deploy(timeout=60)

  # run the program for 120 seconds, then close the session
  session.run(wait=120, close_session=True, close_pipelines=True)
  session.P("Main thread exiting...")



import os
import bz2
import pickle
from time import time

class _PickleSerializationMixin(object):
  """
  Mixin for pickle serialization functionalities that are attached to `ratio1.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `ratio1.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_PickleSerializationMixin, self).__init__()
    return

  def _save_compressed_pickle(self, full_filename, myobj, locking=False):
    """
    save object to file using pickle

    @param full_filename: name of destination file
    @param myobj: object to save (has to be pickleable)
    """
    with self.managed_lock_resource(full_filename, condition=locking):
      try:
        fhandle = bz2.BZ2File(full_filename, 'wb')
        pickle.dump(myobj, fhandle, protocol=pickle.HIGHEST_PROTOCOL)
        fhandle.close()
      except:
        self.P('ERROR: File ' + full_filename + ' cannot be written!')
        return False
    # endwith conditional lock
    return True


  def _load_compressed_pickle(self, full_filename, locking=False):
    """
    Load from filename using pickle

    @param full_filename: name of file to load from
    """
    with self.managed_lock_resource(full_filename, condition=locking):
      try:
        fhandle = bz2.BZ2File(full_filename, 'rb')
        myobj = pickle.load(fhandle)
        fhandle.close()
      except:
        self.P('ERROR: File ' + full_filename + ' cannot be read!')
        return None
    # endwith conditional lock

    return myobj


  def save_pickle(self, data, fn, folder=None,
                  use_prefix=False, verbose=True,
                  compressed=False,
                  subfolder_path=None,
                  locking=True,
                  ):
    """
    compressed: True if compression is required OR you can just add '.pklz' to `fn`
    """

    def P(s):
      if verbose:
        self.P(s)
      return

    # enddef

    lfld = self.get_target_folder(target=folder)

    if lfld is None:
      P("Assuming `fn` param ({}) is a full path".format(fn))
      datafile = fn
    else:
      if use_prefix:
        fn = self.file_prefix + '_' + fn
      datafolder = lfld
      if subfolder_path is not None:
        datafolder = os.path.join(datafolder, subfolder_path.lstrip('/'))
        os.makedirs(datafolder, exist_ok=True)
      datafile = os.path.join(datafolder, fn)

    os.makedirs(os.path.split(datafile)[0], exist_ok=True)

    tm_start = time()
    if compressed or '.pklz' in fn:
      if not compressed:
        P("Saving pickle with compression=True forced due to extension")
      else:
        P("Saving pickle with compression...")
      if self._save_compressed_pickle(datafile, myobj=data, locking=locking):
        tm_elapsed = time() - tm_start
        P("  Compressed pickle {} saved in {} folder in {:.1f}s".format(fn, folder, tm_elapsed))
      else:
        P("  FAILED compressed pickle save!")
    else:
      P("Saving uncompressed pikle (lock:{}) : {} ".format(locking, datafile))
      with self.managed_lock_resource(datafile, condition=locking):
        try:
          with open(datafile, 'wb') as fhandle:
            pickle.dump(data, fhandle, protocol=pickle.HIGHEST_PROTOCOL)
          tm_elapsed = time() - tm_start
        except:
          pass
    # endwith conditional lock
      if verbose:
        P("  Saved pickle '{}' in '{}' folder in {:.1f}s".format(fn, folder, tm_elapsed))
    return datafile


  def save_pickle_to_data(self, data, fn, compressed=False, verbose=True, 
                          subfolder_path=None, locking=True):
    """
    compressed: True if compression is required OR you can just add '.pklz' to `fn`
    """
    return self.save_pickle(
      data, fn,
      folder='data',
      compressed=compressed,
      subfolder_path=subfolder_path,
      verbose=verbose,
      locking=locking,
    )


  def save_pickle_to_models(self, data, fn, compressed=False, verbose=True, 
                            subfolder_path=None, locking=True):
    """
    compressed: True if compression is required OR you can just add '.pklz' to `fn`
    """
    return self.save_pickle(
      data, fn,
      folder='models',
      compressed=compressed,
      subfolder_path=subfolder_path,
      verbose=verbose,
      locking=locking,
    )


  def save_pickle_to_output(self, data, fn, compressed=False, verbose=True, 
                            subfolder_path=None, locking=True):
    """
    compressed: True if compression is required OR you can just add '.pklz' to `fn`
    """
    return self.save_pickle(
      data, fn,
      folder='output',
      compressed=compressed,
      subfolder_path=subfolder_path,
      verbose=verbose,
      locking=locking,
    )


  def load_pickle_from_models(self, fn, decompress=False, verbose=True, 
                              subfolder_path=None, locking=True):
    """
     decompressed : True if the file was saved with `compressed=True` or you can just use '.pklz'
    """
    return self.load_pickle(
      fn,
      folder='models',
      decompress=decompress,
      verbose=verbose,
      subfolder_path=subfolder_path,
      locking=locking,
    )


  def load_pickle_from_data(self, fn, decompress=False, verbose=True, 
                            subfolder_path=None, locking=True):
    """
     decompressed : True if the file was saved with `compressed=True` or you can just use '.pklz'
    """
    return self.load_pickle(
      fn,
      folder='data',
      decompress=decompress,
      verbose=verbose,
      subfolder_path=subfolder_path,
      locking=locking,
    )


  def load_pickle_from_output(self, fn, decompress=False, verbose=True, 
                              subfolder_path=None, locking=True):
    """
     decompressed : True if the file was saved with `compressed=True` or you can just use '.pklz'
    """
    return self.load_pickle(
      fn,
      folder='output',
      decompress=decompress,
      verbose=verbose,
      subfolder_path=subfolder_path,
      locking=locking,
    )


  def load_pickle(self, fn, folder=None, decompress=False, verbose=True,
                  subfolder_path=None, locking=True):
    """
     load_from: 'data', 'output', 'models'
     decompressed : True if the file was saved with `compressed=True` or you can just use '.pklz'
    """
    if verbose:
      P = self.P
    else:
      P = lambda x, color=None: x

    lfld = self.get_target_folder(target=folder)

    if lfld is None:
      P("Loading pickle ... Assuming `fn` param ({}) is a full path".format(fn))
      datafile = fn
    else:
      datafolder = lfld
      if subfolder_path is not None:
        datafolder = os.path.join(datafolder, subfolder_path.lstrip('/'))
      datafile = os.path.join(datafolder, fn)
      P("Loading pickle (locked:{}) from {}".format(locking, datafile))
    #endif full path or not
    data = None
    exc = None
    if os.path.isfile(datafile):
      if decompress or '.pklz' in datafile:
        if not decompress:
          P("Loading pickle with decompress=True forced due to extension")
        else:
          P("Loading pickle with decompression...")
        data = self._load_compressed_pickle(datafile)
      else:
        with self.managed_lock_resource(datafile, condition=locking):
          try:
            with open(datafile, "rb") as f:
              data = pickle.load(f)
          except Exception as _exc:
            data = None
            exc = _exc
        # endwith conditional lock
      #endif decompress or not
      if data is None:
        P("  {} load failed with error {}".format(datafile, exc), color='r')
      else:
        P("  Loaded: {}".format(datafile))
      #endif data is None
    else:
      P("  File not found! Pickle load failed.", color='r')
    return data
  
  
  def update_pickle_from_data(self, 
                              fn, 
                              update_callback, 
                              decompress=False, 
                              verbose=False, 
                              subfolder_path=None,
                              force_update=False):
    assert update_callback is not None, "update_callback must be defined!"
    datafile = self.get_file_path(
      fn=fn,
      folder='data',
      subfolder_path=subfolder_path,
      force=True
      )
    if datafile is None:
      self.P("update_pickle_from_data failed due to missing {}".format(datafile), color='error')
      return False

    with self.managed_lock_resource(datafile):
      result = None
      try:
        data = self.load_pickle_from_data(
          fn=fn,
          decompress=decompress,
          verbose=verbose,
          subfolder_path=subfolder_path,
          locking=False,
          )
        
        if data is not None or force_update:
          data = update_callback(data)
          
          self.save_pickle_to_data(
            data=data, 
            fn=fn,
            compressed=decompress,
            verbose=verbose,
            subfolder_path=subfolder_path,
            locking=False,
            )
          result = True
      except Exception as e:
        self.P("update_pickle_from_data failed: {}".format(e), color='error')
        result = False
    # endwith lock
    return result
      